import pandas as pd
import numpy as np


def preprocess_cpc(file):
    """
    CPC processing
    file: path to cpc csv file (string)
    return: processed dataframe
    """
    df = pd.read_csv(file)
    df = df.dropna(axis=0)
    df = df.reset_index(drop=True)
    df["datetime"] = pd.to_datetime(df["date_time"])
    df.replace(0, np.nan, inplace=True) # 0 values are invalid
    df = df.drop(["date_time"], axis=1)
    time_col = df.pop("datetime")
    df.insert(0, "datetime", time_col)
    df = df.rename(
        {"N conc(1/ccm)": "N_conc_cpc (cm-3)", "Pressure (hPa)": "press_cpc (hPa)"},
        axis=1,
    )
    return df



import pandas as pd
import numpy as np

def calculate_height(p0, p1, T0, T1):
    """
    Calculate height based on hydrostatic pressure equation, assuming a uniform layer
    
    p0: pressure at lower level (hPa)
    p1: pressure at upper level (hPa)
    T0: temperature at lower level (K)
    T1: temperature at upper level (K)
    """
    R = 287.05
    g = 9.80665
    height = R / g * ((T0 + T1) / 2 + 273.15) * np.log(p0 / p1)
    return height


def calculate_height_df(df, p, T):
    """
    Advance calculation of height based on hydrostatic pressure equation, 
    assuming mini uniform layer
    df: dataframe containing pressure and temperature columns
    p: pressure column name (string)
    T: temperature column name (string)
    return: dataframe with height column added
    """
    df_height = df.copy()
    df_height.dropna(subset=p, inplace=True)
    height = np.zeros_like(df_height[p])
    height[1:] = calculate_height(
        df_height[p][:-1].values,
        df_height[p][1:].values,
        df_height[T][:-1].values,
        df_height[T][1:].values,
    )
    df_height["height"] = height
    df["height"] = df_height["height"]
    df.replace({"height": np.nan}, 0, inplace=True)
    df["height"] = df["height"].cumsum()
    return df


def preprocess_bme(file):
    """
    BME processing
    file: path to bme csv file (string)
    return: processed dataframe
    """
    df = pd.read_csv(file)
    df = df.dropna(axis=0)
    df = df.reset_index(drop=True)
    df["datetime"] = pd.to_datetime(df["date"] + " " + df["time"])
    df = df.drop(["date", "time"], axis=1)
    time_col = df.pop("datetime")
    df.insert(0, "datetime", time_col)
    df = df.rename(
        {
            "temp_bme": "temp_bme (C)",
            "press_bme": "press_bme (hPa)",
            "rh_bme": "rh_bme (%)",
        },
        axis=1,
    )
    return df


import pandas as pd
import numpy as np
import importlib.resources

pops_binedges = (
    importlib.resources.files("UAVision.bin_edges").joinpath("pops_binedges.txt").read_text()
)
pops_binedges = np.fromstring(pops_binedges, sep="\n")

def preprocess_pops(file, size=None, drop_aux=True):
    """
    POPS processing

    file: path to pops csv file (string)
    size: optional. If None uses bundled pops_binedges (bin edges).
          If array-like is provided it must be the bin edges with length 17.
    drop_aux: bool, if True drop auxiliary columns (default True). If False keep them.
    return: processed dataframe
    """
    df = pd.read_csv(file)
    df = df.dropna(axis=0)
    df = df.reset_index(drop=True)
    df["datetime"] = pd.to_datetime(df["DateTime"], unit="s")
    df = df.set_index("datetime").resample("1s").mean().dropna().reset_index()
    df = df.drop(["DateTime"], axis=1)
    time_col = df.pop("datetime")
    df.insert(0, "datetime", time_col)

    # determine binedges from provided size or bundled pops_binedges
    if size is None:
        binedges = pops_binedges.astype(float)
    elif isinstance(size, (list, tuple, np.ndarray, pd.Series)):
        arr = np.asarray(size, dtype=float)
        if arr.ndim != 1 or arr.size != 17:
            raise ValueError("When providing 'size' as an array it must be a 1-D array of 17 bin edges.")
        binedges = arr
    else:
        raise TypeError("size must be None or an array-like of 17 bin edges")

    dlog_bin = np.log10(binedges[1:]) - np.log10(binedges[:-1])

    pops_binlab = [
        "b0",
        "b1",
        "b2",
        "b3",
        "b4",
        "b5",
        "b6",
        "b7",
        "b8",
        "b9",
        "b10",
        "b11",
        "b12",
        "b13",
        "b14",
        "b15",
    ]
    dndlog_label = ["bin" + str(x) + "_pops (dN/dlogDp)" for x in range(1, 17)]
    conc_label = ["bin" + str(x) + "_pops (cm-3)" for x in range(1, 17)]
    df = df.rename(columns={x:y for x,y in zip(pops_binlab, conc_label)})
    # Calculate concentration cm-3
    df[conc_label] = df[conc_label].div(df[" POPS_Flow"] * 16.6667, axis=0)
    # Calculate dN/dlogDp
    dndlog = df[conc_label].div(dlog_bin, axis=1)
    dndlog.columns = dndlog_label
    df = pd.concat([df, dndlog], axis=1)
        
    if drop_aux:
        df = df.drop(
            [
                " Status",
                " PartCt",
                " BL",
                " BLTH",
                " STD",
                " TofP",
                " PumpFB",
                " LDTemp",
                " LaserFB",
                " LD_Mon",
                " Temp",
                " BatV",
                " Laser_Current",
                " Flow_Set",
                "PumpLife_hrs",
                " BL_Start",
                " TH_Mult",
                " nbins",
                " logmin",
                " logmax",
                " Skip_Save",
                " MinPeakPts",
                "MaxPeakPts",
                " RawPts",
            ],
            axis=1,
            errors="ignore",
        )
    df = df.rename(
        {
            " PartCon": "N_conc_pops (cm-3)",
            " P": "press_pops (hPa)",
            " POPS_Flow": "flow_rate_pops (l/m)",
        },
        axis=1,
    )
    return df


import pandas as pd
import glob
import os
import re
from functools import reduce
import argparse


def merge_sensor_data(dir_in, dir_out):
    dir_in = dir_in.replace("\\", "/") + "/"
    dir_out = dir_out.replace("\\", "/") + "/"

    sub_dir = [f.path for f in os.scandir(dir_in) if f.is_dir()]
    file_types = [".csv", ".txt"]
    for sub_dir_ in sub_dir:
        file_path = []
        for file_type in file_types:
            file_path.extend([x for x in glob.glob(sub_dir_ + "/*" + file_type)])
        file_name = [os.path.basename(x).rsplit(".", 1)[0] for x in file_path]
        instrument_name = [re.sub(r"[\.\-_][0-9]+", "", x) for x in file_name]
        file_summary = pd.DataFrame(
            {
                "file_path": file_path,
                "file_name": file_name,
                "instrument_name": instrument_name,
            }
        )
        data = {}
        for instrument, grp in file_summary.groupby("instrument_name"):
            data[instrument] = pd.concat(
                [
                    pd.read_csv(
                        x,
                        index_col=False,
                        sep=pd.read_csv(
                            x, sep=None, iterator=True, nrows=2
                        )._engine.data.dialect.delimiter,
                    )
                    for x in grp.file_path
                ],
                ignore_index=True,
            )

        for key, val in data.items():
            data[key] = data[key].dropna(axis=0, how="all")
            data[key].columns = data[key].columns.str.replace(" ", "")
            if "datetime" in data[key].columns:
                data[key]["datetime"] = pd.to_datetime(data[key]["datetime"])
            else:
                if "time" not in data[key].columns:
                    data[key]["datetime"] = pd.to_datetime(data[key]["date"])
                    data[key].drop(["date"], axis=1, inplace=True)
                else:
                    data[key]["datetime"] = pd.to_datetime(
                        data[key]["date"] + " " + data[key]["time"]
                    )
                    data[key].drop(["date", "time"], axis=1, inplace=True)
            data[key] = data[key].set_index("datetime").resample("1s").mean()
            data[key] = data[key].reset_index()
            data[key] = data[key].dropna()
            data[key].columns = [
                x + "_" + key if "datetime" not in x else x for x in data[key].columns
            ]

        data_merged = reduce(
            lambda left, right: pd.merge(left, right, on=["datetime"], how="outer"),
            data.values(),
        )
        data_merged = data_merged.set_index("datetime")
        data_merged = data_merged.sort_index()
        data_merged.reset_index(inplace=True)
        data_merged.to_csv(
            dir_out + sub_dir_.split("/")[-1] + "_merged.csv", index=False
        )
    print(f"{len(sub_dir)} folders merged")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Description for arguments")
    parser.add_argument("dir_in", help="Input directory", type=str)
    parser.add_argument("dir_out", help="Output directory", type=str)
    argument = parser.parse_args()

    merge_sensor_data(argument.dir_in, argument.dir_out)

    print("Finished merging files")



from __future__ import annotations

import importlib
import importlib.metadata as _metadata
from typing import Any

# package version (fall back to static when not installed)
try:
    __version__ = _metadata.version("UAVision")
except _metadata.PackageNotFoundError:
    __version__ = "1.0.0"

# submodules provided by this package; they will be imported lazily
_SUBMODULES = {
    "mavic",
    "bme",
    "cpc",
    "mcda",
    "pops"
}

__all__ = list(_SUBMODULES) + ["__version__"]


def __getattr__(name: str) -> Any:
    """
    Lazily import and expose submodules as attributes on the package.

    Example:
        from UAVision import mcda  # imports UAVision.mcda on demand
    """
    if name in _SUBMODULES:
        module = importlib.import_module(f"{__name__}.{name}")
        globals()[name] = module  # cache on the package module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__() -> list[str]:
    # include lazy submodules in attribute listings (help(), tab-completion)
    return sorted(list(globals().keys()) + list(_SUBMODULES))


import numpy as np
import importlib.resources
import pandas as pd

n2_binedges = (
    importlib.resources.files("UAVision.bin_edges")
    .joinpath("opcN2_binedges.txt")
    .read_text()
)
n2_binedges = np.fromstring(n2_binedges, sep="\n")

n3_binedges = (
    importlib.resources.files("UAVision.bin_edges")
    .joinpath("opcN3_binedges.txt")
    .read_text()
)
n3_binedges = np.fromstring(n3_binedges, sep="\n")


def calculate_concentration(df, bin_label, flow_label=None, period_label=None):
    """
    Calculate dN/dLogDp from OPC N2 and N3
    df: dataframe containing bin counts and flow rate
    bin_label: list of bin column names (string)
    flow_label: flow rate column name (string), only for OPC N3
    period_label: sampling period column name (string), only for OPC N3
    return: dN/dLogDp dataframe for OPC N2
            concentration and dN/dLogDp dataframe for OPC N3
    """
    if len(bin_label) == 16:
        print("OPC-N2")
        dlog_bin = np.log10(n2_binedges[1:]) - np.log10(n2_binedges[:-1])
        dndlogdp = df[bin_label].div(dlog_bin, axis=1)
        return dndlogdp

    elif len(bin_label) == 24:
        print("OPC-N3")
        dlog_bin = np.log10(n3_binedges[1:]) - np.log10(n3_binedges[:-1])
        total_volume = df[flow_label] / 100
        period = df[period_label] / 100
        concentration = df[bin_label].div(total_volume, axis=0).div(period, axis=0)
        dndlogdp = concentration.div(dlog_bin, axis=1)
        return concentration, dndlogdp

def calculate_lag(df, var1, var2, lag):
    """
    Calculate the lag between two variables, same dataframe.
    Please resample the dataframe so spacing is consistent
    df: dataframe containing the two variables
    var1: first variable column name (string)
    var2: second variable column name (string)
    lag: maximum lag to consider (int)
    """
    lag_range = np.arange(-lag, lag+1)
    df_corr = pd.DataFrame({
        'covariance': np.array([df[var1].corr(df[var2].shift(x)) for x in lag_range]),
        'lag': lag_range})
    df_corr['covariance'] = df_corr['covariance'].abs()
    imax = df_corr['covariance'].idxmax()
    lag_max = df_corr['lag'][imax]
    print(f"Max correlation when shift forward {var2} by {lag_max} units")
    return lag_max


import pandas as pd
import glob
import os
import argparse


def merge_wind_data(dir_in, dir_out):
    dir_in = dir_in.replace("\\", "/") + "/"
    dir_out = dir_out.replace("\\", "/") + "/"
    file_path_wind = [x for x in glob.glob(dir_in + "/*.csv")]
    df = pd.DataFrame({})    
    for file in file_path_wind:
        file_name = os.path.basename(file)
        start_time = pd.to_datetime(file_name[-23:-4], format='%Y-%m-%d_%H-%M-%S')
        df_ = pd.read_csv(file)
        df_['datetime'] = start_time + pd.to_timedelta(df_['Flight time'])
        df = pd.concat([df, df_], ignore_index=True)
    df.to_csv(dir_out + "wind_merged.csv", index=False)
    print(f"{len(file_path_wind)} files merged")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Description for arguments")
    parser.add_argument("dir_in", help="Input directory", type=str)
    parser.add_argument("dir_out", help="Output directory", type=str)
    argument = parser.parse_args()

    merge_wind_data(argument.dir_in, argument.dir_out)

    print("Finished merging files")



import importlib.resources
import json
import numpy as np
import pandas as pd

mcda_midbin_all = (
    importlib.resources.files("UAVision.bin_edges").joinpath("mcda_midbin_all.txt").read_text()
)
mcda_midbin_all = json.loads(mcda_midbin_all)


def preprocess_mcda(file, size):
    """
    mCDA processing, calculate derived parameters as well
    file: path to mcda csv file (string)
    size: size category string, one of the following
      ['PSL_0.6-40', 'PSL_0.15-17', 'water_0.6-40', 'water_0.15-17'] OR
      an array-like of mid-bin values (list/tuple/ndarray)
      If an array-like is provided, it must be length 256.
    return: processed dataframe
    """
    # accept an array-like of mid_bin values as well as a size key string
    if isinstance(size, (list, tuple, np.ndarray, pd.Series)):
        mid_bin = np.asarray(size, dtype=float)
        if mid_bin.ndim != 1 or mid_bin.size != 256:
            raise ValueError(
                "When providing an array-like 'size', it must be one-dimensional with length 256."
            )
    elif isinstance(size, str):
        if size not in mcda_midbin_all:
            raise KeyError(
                f"size '{size}' not found. Valid keys: {list(mcda_midbin_all.keys())}"
            )
        mid_bin = np.array(mcda_midbin_all[size], dtype=float)
    else:
        raise TypeError("size must be a key string or an array-like of 256 mid-bin values")

    # calculate size dlog_bin
    print(size)
    binedges = np.append(
        np.append(
            mid_bin[0] - (-mid_bin[0] + mid_bin[1]) / 2,
            (mid_bin[:-1] + mid_bin[1:]) / 2,
        ),
        (mid_bin[-1] - mid_bin[-2]) / 2 + mid_bin[-1],
    )
    dlog_bin = np.log10(binedges[1:]) - np.log10(binedges[:-1])

    # Load file
    df = pd.read_csv(file, skiprows=1, header=None, dtype=str)
    df = df.iloc[:, np.r_[0:257, -6:0]]
    df = df.dropna(axis=0)
    df = df.reset_index(drop=True)
    df.columns = np.arange(df.columns.size)
    df[0] = pd.to_datetime(df[0], format="%Y%m%d%H%M%S")

    dndlog_label = ["bin" + str(x) + "_mcda (dN/dlogDp)" for x in range(1, 257)]
    conc_label = ["bin" + str(x) + "_mcda (cm-3)" for x in range(1, 257)]
    pm_label = [
        "pcount_mcda",
        "pm1_mcda",
        "pm25_mcda",
        "pm4_mcda",
        "pm10_mcda",
        "pmtot_mcda",
    ]
    df.columns = np.r_[["datetime"], conc_label, pm_label]

    # Convert hex to int
    df[conc_label] = df[conc_label].map(
        lambda x: int(x, base=16)
    )
    # Convert to float
    df = df.set_index("datetime").astype("float").reset_index() 
    # Bin counts
    df_bins = df[conc_label].copy().to_numpy().astype(float)
    # Calculate concentration cm-3
    df[conc_label] = df[conc_label] / 10 / 46.67  # 10s averaged, 2.8L/min flow = 46.67 ccm/s

    # Calculate dN/dlogDp
    dndlog = df[conc_label] / dlog_bin
    dndlog.columns = dndlog_label
    df = pd.concat([df, dndlog], axis=1)    
    # Calculate CDNC
    df["Nd_mcda (cm-3)"] = df_bins.sum(axis=1) / 10 / 46.67
    # Calculate LWC
    conc_perbin = df_bins / 10 / (2.8e-3 / 60)
    lwc_perbin = conc_perbin * 1e6 * np.pi / 6 * (mid_bin * 1e-6) ** 3
    lwc_sum = lwc_perbin.sum(axis=1)
    df["LWC_mcda (g/m3)"] = lwc_sum
    # Calculate MVD
    p_lwc_perbin = np.divide(
        lwc_perbin,
        lwc_sum[:, np.newaxis],
        out=np.zeros_like(lwc_perbin),
        where=lwc_sum[:, np.newaxis] != 0,
    )
    cumsum_lwc_perbin = p_lwc_perbin.cumsum(axis=1)
    cumsum_lwc_perbin[df_bins == 0] = np.nan
    # find imin and imax, they contain the point where cumsum_lwc_perbin == 0.5
    imax = np.argmax((cumsum_lwc_perbin > 0.5), axis=1)
    imin = (
        cumsum_lwc_perbin.shape[1]
        - np.argmax(cumsum_lwc_perbin[:, ::-1] < 0.5, axis=1)
        - 1
    )
    # The MVD formula is based on this where max is first non-zero bin cumsum > 0.5 and
    # min is last non-zero bin cumsum < 0.5
    # (0.5 - cum_min) / (cum_max - cum_min) = (bx - bmin) / (bmax - bmin)
    cum_min = cumsum_lwc_perbin[np.arange(len(cumsum_lwc_perbin)), imin]
    cum_max = cumsum_lwc_perbin[np.arange(len(cumsum_lwc_perbin)), imax]
    bmin = mid_bin[imin]
    bmax = mid_bin[imax]
    df["MVD_mcda (um)"] = bmin + (0.5 - cum_min) / (cum_max - cum_min) * (bmax - bmin)
    # Calculate ED
    top = (conc_perbin * mid_bin**3).sum(axis=1)
    bottom = (conc_perbin * mid_bin**2).sum(axis=1)
    df["ED_mcda (um)"] = np.divide(
        top, bottom, out=np.zeros_like(top), where=bottom != 0
    )
    # Drop columns
    df = df.drop(["pcount_mcda", "pm4_mcda", "pmtot_mcda"], axis=1)
    return df


def cloudmask(df):
    import re
    """cloud mask for mcda"""
    if df["datetime"][0] < pd.Timestamp("20221003"):
        size = "water_0.15-17"
    else:
        size = "water_0.6-40"
    cda_midbin = np.array(mcda_midbin_all[size], dtype=float)
    cda_midbin = cda_midbin[81:]
    # RH > 80%
    rh_cloud = df["rh_bme (%)"] > 80
    # Count > 1 in 10s with size > 2 um
    bin_lab = [x for x in df.columns if re.search(r"bin[0-9]+_mcda \(cm-3\)", x)]
    bin_lab_cloud = bin_lab[np.argmax(cda_midbin > 2) :]
    bin_count = df[bin_lab]
    count_cloud = bin_count[bin_lab_cloud].sum(axis=1) * 10 > 5
    cloudmask = rh_cloud & count_cloud
    return cloudmask



import numpy as np

def calculate_binedges(midbin):
    """
    calculate bin edges from mid bin
    midbin: mid bin array
    return: binedges array
    """
    binedges = np.append(
        np.append(
            midbin[0] - (-midbin[0] + midbin[1]) / 2, (midbin[:-1] + midbin[1:]) / 2
        ),
        (midbin[-1] - midbin[-2]) / 2 + midbin[-1],
    )
    return binedges


def calculate_midbin(binedges):
    """
    calculate midbin from bin edges
    binedges: binedges array
    return: midbin array
    """
    midbin = (binedges[1:] + binedges[:-1]) / 2
    return midbin



from __future__ import annotations

import importlib
from typing import Any

# submodules provided by this package; they will be imported lazily
_SUBMODULES = {
    "merge_sensor_data",
    "merge_wind_data",
    "preprocess",
}

__all__ = list(_SUBMODULES) + ["__version__"]


def __getattr__(name: str) -> Any:
    """
    Lazily import and expose submodules as attributes on the package.

    Example:
        from UAVision import mcda  # imports UAVision.mcda on demand
    """
    if name in _SUBMODULES:
        module = importlib.import_module(f"{__name__}.{name}")
        globals()[name] = module  # cache on the package module
        return module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
